{% import 'keg-elements/forms/horizontal-b4.html' as horizontal %}
{% extends 'keg-elements/form-view.html' %}
{%- if _ is not defined -%}
    {% from 'keg-auth/i18n.j2' import gettext as _ %}
{%- endif -%}

{% block form %}
{{
    horizontal.form(form, action_text=_('Send Reset Email'),
                    cancel_url=auth_manager.url_for('login'))
}}
{% endblock %}
